using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.PluginInfrastructure;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Random;

namespace HeuristicLab.Algorithms.DataAnalysis {
  [StorableClass]
  [Creatable(CreatableAttribute.Categories.DataAnalysisRegression, Priority = 95)]
  [Item("M5RegressionTree", "A M5 regression tree / rule set")]
  public sealed class M5Regression : FixedDataAnalysisAlgorithm<IRegressionProblem> {
    #region Parametername
    private const string GenerateRulesParameterName = "GenerateRules";
    private const string HoldoutSizeParameterName = "HoldoutSize";
    private const string SpliterParameterName = "Spliter";
    private const string MinimalNodeSizeParameterName = "MinimalNodeSize";
    private const string LeafModelParameterName = "LeafModel";
    private const string PruningTypeParameterName = "PruningType";
    private const string SeedParameterName = "Seed";
    private const string SetSeedRandomlyParameterName = "SetSeedRandomly";
    private const string UseHoldoutParameterName = "UseHoldout";
    #endregion

    #region Parameter properties
    public IFixedValueParameter<BoolValue> GenerateRulesParameter {
      get { return (IFixedValueParameter<BoolValue>)Parameters[GenerateRulesParameterName]; }
    }
    public IFixedValueParameter<PercentValue> HoldoutSizeParameter {
      get { return (IFixedValueParameter<PercentValue>)Parameters[HoldoutSizeParameterName]; }
    }
    public IConstrainedValueParameter<ISpliter> ImpurityParameter {
      get { return (IConstrainedValueParameter<ISpliter>)Parameters[SpliterParameterName]; }
    }
    public IFixedValueParameter<IntValue> MinimalNodeSizeParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[MinimalNodeSizeParameterName]; }
    }
    public IConstrainedValueParameter<ILeafModel> LeafModelParameter {
      get { return (IConstrainedValueParameter<ILeafModel>)Parameters[LeafModelParameterName]; }
    }
    public IConstrainedValueParameter<IPruning> PruningTypeParameter {
      get { return (IConstrainedValueParameter<IPruning>)Parameters[PruningTypeParameterName]; }
    }
    public IFixedValueParameter<IntValue> SeedParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[SeedParameterName]; }
    }
    public IFixedValueParameter<BoolValue> SetSeedRandomlyParameter {
      get { return (IFixedValueParameter<BoolValue>)Parameters[SetSeedRandomlyParameterName]; }
    }
    public IFixedValueParameter<BoolValue> UseHoldoutParameter {
      get { return (IFixedValueParameter<BoolValue>)Parameters[UseHoldoutParameterName]; }
    }
    #endregion

    #region Properties 
    public bool GenerateRules {
      get { return GenerateRulesParameter.Value.Value; }
    }
    public double HoldoutSize {
      get { return HoldoutSizeParameter.Value.Value; }
    }
    public ISpliter Split {
      get { return ImpurityParameter.Value; }
    }
    public int MinimalNodeSize {
      get { return MinimalNodeSizeParameter.Value.Value; }
    }
    public ILeafModel LeafModel {
      get { return LeafModelParameter.Value; }
    }
    public IPruning Pruning {
      get { return PruningTypeParameter.Value; }
    }
    public int Seed {
      get { return SeedParameter.Value.Value; }
    }
    public bool SetSeedRandomly {
      get { return SetSeedRandomlyParameter.Value.Value; }
    }
    public bool UseHoldout {
      get { return UseHoldoutParameter.Value.Value; }
    }
    #endregion

    #region Constructors and Cloning
    [StorableConstructor]
    private M5Regression(bool deserializing) : base(deserializing) { }
    private M5Regression(M5Regression original, Cloner cloner) : base(original, cloner) { }
    public M5Regression() {
      var modelSet = new ItemSet<ILeafModel>(ApplicationManager.Manager.GetInstances<ILeafModel>());
      var pruningSet = new ItemSet<IPruning>(ApplicationManager.Manager.GetInstances<IPruning>());
      var impuritySet = new ItemSet<ISpliter>(ApplicationManager.Manager.GetInstances<ISpliter>());
      Parameters.Add(new FixedValueParameter<BoolValue>(GenerateRulesParameterName, "Whether a set of rules or a decision tree shall be created", new BoolValue(false)));
      Parameters.Add(new FixedValueParameter<PercentValue>(HoldoutSizeParameterName, "How much of the training set shall be reserved for pruning", new PercentValue(0.2)));
      Parameters.Add(new ConstrainedValueParameter<ISpliter>(SpliterParameterName, "The type of split function used to create node splits", impuritySet, impuritySet.OfType<M5Spliter>().First()));
      Parameters.Add(new FixedValueParameter<IntValue>(MinimalNodeSizeParameterName, "The minimal number of samples in a leaf node", new IntValue(1)));
      Parameters.Add(new ConstrainedValueParameter<ILeafModel>(LeafModelParameterName, "The type of model used for the nodes", modelSet, modelSet.OfType<LinearLeaf>().First()));
      Parameters.Add(new ConstrainedValueParameter<IPruning>(PruningTypeParameterName, "The type of pruning used", pruningSet, pruningSet.OfType<M5LinearBottomUpPruning>().First()));
      Parameters.Add(new FixedValueParameter<IntValue>(SeedParameterName, "The random seed used to initialize the new pseudo random number generator.", new IntValue(0)));
      Parameters.Add(new FixedValueParameter<BoolValue>(SetSeedRandomlyParameterName, "True if the random seed should be set to a random value, otherwise false.", new BoolValue(true)));
      Parameters.Add(new FixedValueParameter<BoolValue>(UseHoldoutParameterName, "True if a holdout set should be generated, false if splitting and pruning shall be performed on the same data ", new BoolValue(false)));
      Problem = new RegressionProblem();
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new M5Regression(this, cloner);
    }
    #endregion

    protected override void Run(CancellationToken cancellationToken) {
      var random = new MersenneTwister();
      if (SetSeedRandomly) SeedParameter.Value.Value = new System.Random().Next();
      random.Reset(Seed);
      var solution = CreateM5RegressionSolution(Problem.ProblemData, random, LeafModel, Split, Pruning, UseHoldout, HoldoutSize, MinimalNodeSize, GenerateRules, Results, cancellationToken);
      AnalyzeSolution(solution);
    }

    #region Static Interface
    public static IRegressionSolution CreateM5RegressionSolution(IRegressionProblemData problemData, IRandom random,
      ILeafModel leafModel = null, ISpliter spliter = null, IPruning pruning = null,
      bool useHoldout = false, double holdoutSize = 0.2, int minNumInstances = 4, bool generateRules = false, ResultCollection results = null, CancellationToken? cancellationToken = null) {
      //set default values
      if (leafModel == null) leafModel = new LinearLeaf();
      if (spliter == null) spliter = new M5Spliter();
      if (cancellationToken == null) cancellationToken = CancellationToken.None;
      if (pruning == null) pruning = new M5LeafBottomUpPruning();

      var doubleVars = new HashSet<string>(problemData.Dataset.DoubleVariables);
      var vars = problemData.AllowedInputVariables.Concat(new[] {problemData.TargetVariable}).ToArray();
      if (vars.Any(v => !doubleVars.Contains(v))) throw new NotSupportedException("M5 regression supports only double valued input or output features.");

      var values = vars.Select(v => problemData.Dataset.GetDoubleValues(v, problemData.TrainingIndices).ToArray()).ToArray();
      if (values.Any(v => v.Any(x => double.IsNaN(x) || double.IsInfinity(x))))
        throw new NotSupportedException("M5 regression does not support NaN or infinity values in the input dataset.");

      var trainingData = new Dataset(vars, values);
      var pd = new RegressionProblemData(trainingData, problemData.AllowedInputVariables, problemData.TargetVariable);
      pd.TrainingPartition.End = pd.TestPartition.Start = pd.TestPartition.End = pd.Dataset.Rows;
      pd.TrainingPartition.Start = 0;

      //create & build Model
      var m5Params = new M5Parameters(pruning, minNumInstances, leafModel, pd, random, spliter, results);

      IReadOnlyList<int> trainingRows, pruningRows;
      GeneratePruningSet(problemData.TrainingIndices.ToArray(), random, useHoldout, holdoutSize, out trainingRows, out pruningRows);

      IM5Model model;
      if (generateRules)
        model = M5RuleSetModel.CreateRuleModel(problemData.TargetVariable, m5Params);
      else
        model = M5TreeModel.CreateTreeModel(problemData.TargetVariable, m5Params);

      model.Build(trainingRows, pruningRows, m5Params, cancellationToken.Value);
      return model.CreateRegressionSolution(problemData);
    }

    public static void UpdateM5Model(IRegressionModel model, IRegressionProblemData problemData, IRandom random,
      ILeafModel leafModel, CancellationToken? cancellationToken = null) {
      var m5Model = model as IM5Model;
      if (m5Model == null) throw new ArgumentException("This type of model can not be updated");
      UpdateM5Model(m5Model, problemData, random, leafModel, cancellationToken);
    }

    private static void UpdateM5Model(IM5Model model, IRegressionProblemData problemData, IRandom random,
      ILeafModel leafModel = null, CancellationToken? cancellationToken = null) {
      if (cancellationToken == null) cancellationToken = CancellationToken.None;
      var m5Params = new M5Parameters(leafModel, problemData, random);
      model.Update(problemData.TrainingIndices.ToList(), m5Params, cancellationToken.Value);
    }
    #endregion

    #region Helpers
    private static void GeneratePruningSet(IReadOnlyList<int> allrows, IRandom random, bool useHoldout, double holdoutSize, out IReadOnlyList<int> training, out IReadOnlyList<int> pruning) {
      if (!useHoldout) {
        training = allrows;
        pruning = allrows;
        return;
      }
      var perm = new Permutation(PermutationTypes.Absolute, allrows.Count, random);
      var cut = (int)(holdoutSize * allrows.Count);
      pruning = perm.Take(cut).Select(i => allrows[i]).ToArray();
      training = perm.Take(cut).Select(i => allrows[i]).ToArray();
    }

    private void AnalyzeSolution(IRegressionSolution solution) {
      Results.Add(new Result("RegressionSolution", (IItem)solution.Clone()));

      Dictionary<string, int> frequencies;
      if (!GenerateRules) {
        Results.Add(M5Analyzer.CreateLeafDepthHistogram((M5TreeModel)solution.Model));
        frequencies = M5Analyzer.GetTreeVariableFrequences((M5TreeModel)solution.Model);
      }
      else {
        Results.Add(M5Analyzer.CreateRulesResult((M5RuleSetModel)solution.Model, Problem.ProblemData, "M5TreeResult", true));
        frequencies = M5Analyzer.GetRuleVariableFrequences((M5RuleSetModel)solution.Model);
        Results.Add(M5Analyzer.CreateCoverageDiagram((M5RuleSetModel)solution.Model, Problem.ProblemData));
      }

      //Variable frequencies
      var sum = frequencies.Values.Sum();
      sum = sum == 0 ? 1 : sum;
      var impactArray = new DoubleArray(frequencies.Select(i => (double)i.Value / sum).ToArray()) {
        ElementNames = frequencies.Select(i => i.Key)
      };
      Results.Add(new Result("Variable Frequences", "relative frequencies of variables in rules and tree nodes", impactArray));
    }
    #endregion
  }
}